from Data.Generator import populate, initialize_voyage_fields
from V1.AsiaToEuropeProblem import AsiaToEuropeProblem
from random import shuffle, randint, sample, uniform
from Saver import load_instance, save_instance
from time import time
from VesselPoolGenerator import load_vessel_pool
import os

first_port_cost = 1000

# HELPERS

def distance(node, port_one, port_two):
    return node.distance_matrix[(port_one, port_two)]

def get_time(node, port_one, port_two, speed):
    distance_value = distance(node, port_one, port_two)
    return distance_value/(speed*24)

def get_route_distance(route, node):
    route_distance = 0;
    last_port = None
    for (p, visit) in enumerate(route):

        if visit[0]:
            next_port = p

            if last_port != None:
                route_distance += distance(node, last_port, next_port)
            else:
                route_distance += first_port_cost

            last_port = next_port

    return route_distance

def shuffle_sorted_costs(sorted_costs):
    sorted_costs_shuffled = []

    if len(sorted_costs) > 0:
        last_key = 0
        value = sorted_costs[0][2]
        for (next_key, element) in enumerate(sorted_costs):
            if element[2] != value:
                value = element[2]

                slice = sorted_costs[last_key:next_key]
                shuffle(slice)
                sorted_costs_shuffled += slice
                last_key = next_key

            if next_key == len(sorted_costs)-1:
                slice = sorted_costs[last_key:]
                shuffle(slice)
                sorted_costs_shuffled += slice
                return sorted_costs_shuffled

    else:
        return sorted_costs

# CHECK FREQUENCY CONSTRAINTS

def check_frequency_constraint(node, contract_key, demand_left, max_loading):
    return (node.frequency_counter[contract_key] <= node.max_contract_frequency[contract_key] - node.min_contract_frequency[contract_key] + 1)
    #return (node.frequency_counter[contract_key] >= node.min_contract_frequency[contract_key] and max_loading <= demand_left) and node.frequency_counter[contract_key] - 1 >= node.min_contract_frequency[contract_key]

# CHECK TIME CONSTRAINTS

def check_time_costraint(node, voyage_contracts, voyage_visits):
    voyage_time = [0]*node.ports

    _adjust_visit_times(node, voyage_time, voyage_visits)

    for (contract_key, (load_port, unload_port, transit_time, product)) in enumerate(node.contracts_info):
        if voyage_contracts[contract_key]:
            end_time = voyage_time[unload_port]
            start_time = voyage_time[load_port]
            delta_time = end_time - start_time

            if delta_time > transit_time:
                if not _adjust_time(contract_key, voyage_visits, node):
                    return False
                else:
                    _adjust_visit_times(node, voyage_time, voyage_visits)

    return True

def _adjust_time(contract, voyage_visits, node):

    load_port, unload_port, transit_time, product = node.contracts_info[contract]

    route = voyage_visits[load_port-1: unload_port]

    last_port = None;
    route_distance = 0;
    for (p, visit) in enumerate(route):

        if visit[0]:
            next_port = p

            if last_port != None:
                route_distance += distance(node, last_port, next_port)

            last_port = next_port

    optimal_speed = max(route_distance/(24*transit_time), node.speed[0])

    if optimal_speed > node.speed[-1]:
        return False

    for (key, visit) in enumerate(route[:-1]):
        if visit[1] != None:
            voyage_visit_index = load_port + key - 1
            voyage_visits[voyage_visit_index] = (1, optimal_speed)

    return True

def _adjust_visit_times(node, voyage_time, voyage_visits):
    last_port = None;
    for (p, visit) in enumerate(voyage_visits):

        if visit[0]:
            next_port = p

            if last_port != None:
                voyage_time[p] = get_time(node, last_port, next_port, visit[1]) + voyage_time[last_port]

            last_port = next_port

# CHECK DEMAND CONSTRAINTS

def check_demand_constraints(node):

    for contract_key in range(node.contracts):

        if not _check_demand_for_contract(contract_key, node):
            return False

    return True

def _check_demand_for_contract(contract, node):
    demand = node.contracts_load[contract][0]
    fulfilled_demand = 0

    for loaded_for_voyage in node.voyage_contract_loaded:
        fulfilled_demand += loaded_for_voyage[contract]

    return demand == fulfilled_demand

# GENERATE INSTANCE

def add_contract(node):

    costs = []

    for (voyage_key, voyage_visit) in enumerate(node.voyage_visits):
        visit_counts = sum([x[0] for x in voyage_visit])
        for (contract_key, is_loading) in enumerate(node.voyage_contracts[voyage_key]):

            if not is_loading and not _check_demand_for_contract(contract_key, node):
                max_loading, demand_left = _get_maximum_loading(contract_key, voyage_key, node)
                #total_demand = node.contracts_demand[contract_key]
                total_demand, _, _ = problem.contracts_load[contract_key]
                critical_index = (demand_left/total_demand)*(node.max_contract_frequency[contract_key] - node.min_contract_frequency[contract_key]-node.frequency_counter[contract_key])

                if max_loading != None:
                    marginal_cost_per_unit = _contract_addition_cost(contract_key, voyage_key, node) # / max_loading
                    sort_value = (marginal_cost_per_unit) / (1 + visit_counts)
                    costs.append((sort_value, critical_index, marginal_cost_per_unit, contract_key, voyage_key, max_loading, visit_counts))

    sorted_costs = sorted(costs)
    sorted_costs_shuffled = sorted_costs

    legal_contract = None
    i = 0
    while legal_contract == None:

        if len(sorted_costs_shuffled) == 0 or len(sorted_costs_shuffled) == i:
            return False

        contract = sorted_costs_shuffled[i][3]
        voyage = sorted_costs_shuffled[i][4]
        load_port, unload_port, transit_time, product = node.contracts_info[contract]

        test_voyage_contracts = node.voyage_contracts[voyage][:]
        test_voyage_contracts[contract] = 1

        test_voyage_visits = node.voyage_visits[voyage]
        test_voyage_visits[load_port] = (1, 10)
        test_voyage_visits[unload_port] = (1, 10)

        if not check_time_costraint(node, test_voyage_contracts, test_voyage_visits):
            i += 1
            continue

        max_loading = sorted_costs[i][5]
        node.frequency_counter[contract] += 1
        node.voyage_contracts[voyage][contract] = 1
        node.voyage_visits[voyage] = test_voyage_visits
        _add_load(contract, node.voyage_loaded[voyage], node.voyage_contract_loaded[voyage], max_loading, node, voyage)

        return True

def add_specific_contract(node, contract_key):

    voyage_key = 0

    load_port, unload_port, transit_time = node.contracts_info[contract_key]

    test_voyage_contracts = node.voyage_contracts[voyage_key][:]
    test_voyage_contracts[contract_key] = 1

    test_voyage_visits = node.voyage_visits[voyage_key]
    test_voyage_visits[load_port] = (1, 10)
    test_voyage_visits[unload_port] = (1, 10)

    max_loading = _get_maximum_loading(contract_key, voyage_key, node)[0]

    if check_time_costraint(node, test_voyage_contracts, test_voyage_visits) and max_loading != None:
        node.frequency_counter[contract_key] += 1
        node.voyage_contracts[voyage_key][contract_key] = 1
        node.voyage_visits[voyage_key] = test_voyage_visits
        _add_load(contract_key, node.voyage_loaded[voyage_key], node.voyage_contract_loaded[voyage_key], max_loading, node, voyage_key)

def _contract_addition_cost(contract_key, voyage_key, node):

    load_port, unload_port, transit_time, product = node.contracts_info[contract_key]

    new_voyage_visits = node.voyage_visits[voyage_key][:]
    new_voyage_visits[load_port] = (1, 10)
    new_voyage_visits[unload_port] = (1, 10)

    marginal_cost =  get_route_distance(new_voyage_visits, node) - get_route_distance(node.voyage_visits[voyage_key], node)

    if problem.voyage_visits[voyage_key][load_port][0] == 1 and problem.voyage_visits[voyage_key][unload_port][0] == 1:
        return 0
    else:
        return marginal_cost

def _get_maximum_loading(contract_key, voyage_key, node):
    voyage_loaded = node.voyage_loaded[voyage_key]
    load_port, unload_port, transit_time, product = node.contracts_info[contract_key]

    demand, min_load, max_load = node.contracts_load[contract_key]
    used_capacity = max(voyage_loaded[load_port:unload_port])
    available_capacity = node.vessel_capacity - used_capacity

    loaded = sum([voyage_contract_loaded[contract_key] for voyage_contract_loaded in node.voyage_contract_loaded])
    demand_left = demand-loaded

    max_loading = min(max_load, available_capacity, demand_left)

    if not check_frequency_constraint(node, contract_key, demand_left, max_loading):

        return (None, demand_left)

    if max_loading >= min_load:
        return (max_loading, demand_left)
    else:
        if _adjust_loaded_for_contract(contract_key, node, min_load, demand_left):
            return (min_load, demand_left)
        else:
            return (None, demand_left)

def _adjust_loaded_for_contract(contract_key, node, min_load, demand_left):

    must_remove = min_load - demand_left
    can_remove = 0

    count = 0.0

    if must_remove < 0:
        return False
    else:

        for (voyage_key, contract_loaded) in enumerate(node.voyage_contract_loaded):
            if node.voyage_contract_loaded[voyage_key][contract_key]:
                count += 1.0
                can_remove  += contract_loaded[contract_key] - min_load

        if can_remove >= must_remove and count > 0.0:
            average = must_remove/count
            for (voyage_key, contract_loaded) in enumerate(node.voyage_contract_loaded):
                if node.voyage_contract_loaded[voyage_key][contract_key]:
                    _remove_load(contract_key, node.voyage_loaded[voyage_key], contract_loaded, average, node)


            return True

    return False

def _add_load(contract, voyage_loaded, voyage_contract_loaded, load, node, voyage):
    load_port, unload_port, transit_time, product = node.contracts_info[contract]

    for i in range(load_port, unload_port):
        voyage_loaded[i] += load

    voyage_contract_loaded[contract] += load

def _remove_load(contract, voyage_loaded, voyage_contract_loaded, load, node):
    load_port, unload_port, transit_time, product = node.contracts_info[contract]

    for i in range(load_port, unload_port):
        voyage_loaded[i] -= load

    voyage_contract_loaded[contract] -= load

def first_contracts(problem, number_of_contracts):
    contract_keys = [key for (key, x) in enumerate(problem.voyage_contracts[0])]
    return sample(contract_keys, 5)

def get_solutions(problem):

    x = dict()
    delta = dict()
    l = dict()
    q = dict()
    t = dict()
    w = dict()

    for (voyage_key, voyage_contracts) in enumerate(problem.voyage_contracts):
        p = 1
        k = voyage_key + 1
        v = voyage_key + 1

        for (contract_key, contract_loaded) in enumerate(voyage_contracts):
            c = contract_key + 1

            q[(c, v, p, k)] = problem.voyage_contract_loaded[voyage_key][contract_key]

            if contract_loaded:
                delta[(c, v, k)] = 1

        last_port_key = None
        last_port = None
        cumulative_time = 0

        visit_count = sum(x[0] for x in problem.voyage_visits[voyage_key])
        if visit_count:
            for (port_key, visit) in enumerate(problem.voyage_visits[voyage_key]):
                next_port = port_key + problem.voyages + 1

                if visit[0]:

                    if last_port_key == None:
                        t[(v, next_port)] = 0
                        x[(v, next_port, v, v)] = 1
                        last_port_key = port_key
                        last_port = next_port
                    else:
                        cumulative_time += get_time(problem, last_port_key, port_key, visit[1])
                        t[(v, next_port)] = cumulative_time
                        x[(last_port, next_port, v, v)] = 1
                        l[last_port, next_port, v, p] = problem.voyage_loaded[voyage_key][last_port_key]
                        last_port_key = port_key
                        last_port = next_port

            if last_port != None:
                x[(last_port, problem.voyages + problem.ports + v, v, v)] = 1
        else:
            x[(v, problem.voyages + problem.ports + v, v, v)] = 1

    return x, delta, l, q, t, w

def print_problem(problem, dir=None):

    #### PRINT OUT PROBLEM ####

    fileName = "Problem_ID" + str(problem.id) + "_" + problem.case_name + "_" + str(problem.demand_capacity_ratio) + "Ratio_" + str(problem.contracts) + "Contracts_" + str(problem.vessels) + "Vessels_" + str(problem.voyages) + "Voyages_" + str(problem.products) + "Timestamp" + str(time()) + ".txt"

    if dir == None:
        fileName = "Output/" + fileName
    else:
        fileName = "Output/" + dir + "/" + fileName


    with open(fileName, 'w') as file:

        file.write("portCount: " + str(problem.ports) + "\n")
        file.write("vesselCount: " + str(problem.vessels) + "\n")
        file.write("voyageCount: " + str(problem.voyages) + "\n")
        file.write("contractCount: " + str(problem.contracts) + "\n")
        file.write("productCount: " + str(problem.products) + "\n\n")
        file.write("velocitiesCount: " + str(len(problem.speed)) + "\n\n")

        # GENERATE CONTRACT FREQUENCY MATRIX

        contractFrequencyMatrix = dict()

        for contract in range(problem.contracts):
            minContractFrequency = problem.min_contract_frequency[contract]
            maxContractFrequency = problem.max_contract_frequency[contract]
            contractFrequencyMatrix[contract] = (minContractFrequency, maxContractFrequency)

        # VESSEL CAPACITY
        file.write("VesselCapacity: [\n")
        file.write("! (Product, Vessel)\n")

        divider = float(sum([x for x in range(1, problem.products + 1)]))

        for vessel_key in range(problem.vessels):
            for product_key in range(problem.products):

                k = vessel_key + 1
                p = product_key + 1

                capacity = problem.vessel_capacities[vessel_key][product_key]

                file.write("\t(" + str(p) + "," + str(k) + ")\t" + str(capacity) + "\n")

        file.write("]\n\n")

        # SPACE UTILIZATION

        file.write("SpaceUtilization: [\n")
        file.write("! (Product, Vessel)\n")
        for k in range(1, problem.vessels + 1):

            for p in range(1, problem.products + 1):


                file.write("\t(" + str(p) + "," + str(k) + ")\t" + str(problem.space_utilization[k-1][p-1]) + "\n")

        file.write("]\n\n")

        # VESSEL AVAILABLE AT

        file.write("VesselAvailableAt: [\n")

        file.write("! (Vessel)\n")

        for vessel_key in range(problem.vessels):
            k = vessel_key + 1

            file.write("\t(" + str(k) + ")\t" + str(problem.available_at[vessel_key]) + "\n")

        file.write("]\n\n")

        # SAILING TIME

        file.write("SailingTime: [\n")

        for k in range(1, problem.vessels + 1):

            for m, velocity in enumerate(problem.speed):
                file.write("\t(" + str(k) + "," + str(m + 1) + ")\t" + str(1 / (float(velocity) * 24)) + "\n")

        file.write("]\n\n")

        # SAILING COST

        file.write("SailingCost: [\n")

        for k in range(1, problem.vessels + 1):

            for m, velocity in enumerate(problem.speed):
                cost = problem.sailing_costs[k-1]

                file.write("\t(" + str(k) + "," + str(m + 1) + ")\t" + str(cost[m]) + "\n")

        file.write("]\n\n")

        # DISTANCE

        file.write("\nDistance: [\n")

        for i in range(problem.ports):

            fromPortNumber = i + problem.vessels + 1

            for j in range(i + 1, problem.ports):
                toPortNumber = j + problem.vessels + 1

                file.write("\t(" + str(fromPortNumber) + "," + str(toPortNumber) + ") " + str(
                    problem.distance_matrix[(i, j)]) + "\n")
                file.write("\t(" + str(toPortNumber) + "," + str(fromPortNumber) + ") " + str(
                    problem.distance_matrix[(i, j)]) + "\n")

        file.write("\n]\n")


        # CONTRACT DEMAND

        file.write("\nContractDemand: [\n")
        file.write("\t! (Contract,Product) Demand\n")

        c = 0
        for contract_key in range(problem.contracts):
            for product_key in range(problem.products):
                c = contract_key + 1
                p = product_key + 1

                demand, _, _ = problem.contracts_load[contract_key][product_key]

                file.write("\t(" + str(c) + "," + str(p) + ")\t " + str(round(demand)) + "\n")

        file.write("]\n")

        # CONTRACT PORTS
        file.write("\nContractPorts: [\n")
        for load_port_key, unload_port_key, _ in problem.contracts_info:

            load_port = load_port_key + problem.voyages + 1
            unload_port = unload_port_key + problem.voyages + 1

            file.write("\t" + str(load_port) + "\t" + str(unload_port) + "\n")

        file.write("]\n")

        # CONTRACT TRANSIT TIME

        file.write("\nContractTransitTime: [\n")

        for _, _, transit_time in problem.contracts_info:

            file.write("\t" + str(transit_time) + "\n")

        file.write("]\n\n")

        # MINIMUM QUANTITY CONSTRAINT
        file.write("\nMinimumQuantity: [\n")
        c = 0
        for contract_key in range(problem.contracts):
            for product_key in range(problem.products):
                c = contract_key + 1
                p = product_key + 1

                _, min_load, _ = problem.contracts_load[contract_key][product_key]

                file.write("\t(" + str(c) + "," + str(p) + ")\t " + str(min_load) + "\n")

        file.write("]\n\n")

        # MAXIMUM QUANTITY CONSTRAIN
        file.write("\nMaximumQuantity: [\n")
        c = 0
        for contract_key in range(problem.contracts):
            for product_key in range(problem.products):
                c = contract_key + 1
                p = product_key + 1

                _, _, max_load = problem.contracts_load[contract_key][product_key]

                file.write("\t(" + str(c) + "," + str(p) + ")\t " + str(max_load) + "\n")

        file.write("]\n")

        # PORT VISITS

        file.write("\nMinimumPickups: [\n")
        for contractFrequency in contractFrequencyMatrix.values():
            file.write("\t" + str(int(contractFrequency[0])) + "\n")

        file.write("]\n")

        file.write("\nMaximumPickups: [\n")
        for contractFrequency in contractFrequencyMatrix.values():
            file.write("\t" + str(int(contractFrequency[1])) + "\n")

        file.write("]\n\n")

        # EvenlySpreadContracts
        file.write("\nEvenlySpreadContracts: [\n")
        c = 0
        for c in problem.evenly_spread_contracts:

            file.write("\t" + str(c) + "\n")

        file.write("]\n")

        # PortCost
        file.write("\nPortCosts: [\n")
        n = problem.vessels + 1
        for cost in problem.port_costs:

            file.write("\t(" + str(n) + ")\t " + str(cost) + "\n")
            n += 1

        file.write("]\n")

        # SlackCost
        file.write("\nSlackCost: [\n")
        for contract_key, slack_cost in enumerate(problem.slack_costs):
            c = contract_key + 1

            if c in problem.evenly_spread_contracts:
                file.write("\t(" + str(c) + ")\t " + str(slack_cost) + "\n")

        file.write("]\n")

def get_x(problem, all_ports = False):

    x = dict()

    if all_ports:
        for voyage_key in range(problem.voyages):
            for port_key in range(problem.ports):
                problem.voyage_visits[voyage_key][port_key] = 0

            for load_port, unload_port, _ in problem.contracts_info:
                problem.voyage_visits[voyage_key][load_port] = 1
                problem.voyage_visits[voyage_key][unload_port] = 1

    for (voyage_key, voyage_contracts) in enumerate(problem.voyage_contracts):
        v = voyage_key + 1

        last_port_key = None
        last_port = None

        visit_count = sum(x for x in problem.voyage_visits[voyage_key])
        if visit_count:
            for (port_key, visit) in enumerate(problem.voyage_visits[voyage_key]):
                next_port = port_key + problem.voyages + 1

                if visit:

                    if last_port_key == None:
                        x[(v, next_port, v, v)] = 1
                        last_port_key = port_key
                        last_port = next_port
                    else:
                        x[(last_port, next_port, v, v)] = 1
                        last_port_key = port_key
                        last_port = next_port

            if last_port != None:
                x[(last_port, problem.voyages + problem.ports + v, v, v)] = 1
        else:
            x[(v, problem.voyages + problem.ports + v, v, v)] = 1

    return x

def correct_port_frequency(problem):
    new_port_frequency = [int((f/4.0)*problem.voyages) for f in problem.max_port_frequency]
    problem.max_port_frequency = new_port_frequency

def extract_information(problem):

    min = dict()
    max = dict()

    lord_port_contracts = dict()

    for contract_key, contract in enumerate(problem.contracts_info):
        load_port, unload_port, transit_time =  contract

        # MINIMUM
        if min.get(load_port, 0) < problem.min_contract_frequency[contract_key]:
            min[load_port] = problem.min_contract_frequency[contract_key]

        if min.get(unload_port, 0) < problem.min_contract_frequency[contract_key]:
            min[unload_port] = problem.min_contract_frequency[contract_key]

        # MAXIMUM
        if max.get(load_port, 0) < problem.max_contract_frequency[contract_key]:
            max[load_port] = problem.max_contract_frequency[contract_key]

        if max.get(unload_port, 0) < problem.max_contract_frequency[contract_key]:
            max[unload_port] = problem.max_contract_frequency[contract_key]

        # PORTS
        load_port_array = lord_port_contracts.get(load_port, [])

        load_port_array.append(contract_key)

        lord_port_contracts[load_port] = load_port_array

    print(min)
    print(max)

def add_more_vessels(problem, vessel_counts):

    vessel_pool = load_vessel_pool(problem.vessel_pool_file_name)

    start_vessel_count = problem.vessels

    for vessel_count in vessel_counts:

        vessel_diff = start_vessel_count + vessel_count - problem.vessels

        for i in range(vessel_diff):
            vessel_number = randint(1, len(vessel_pool.capacities) + 1)
            capacity = vessel_pool.capacities[vessel_number]
            space_utilization = [vessel_pool.space_utilization[vessel_number, product] for product in range(1, problem.products + 1)]
            fuel_costs = vessel_pool.fuel_costs[vessel_number]
            problem.vessel_capacities.append(capacity)
            problem.space_utilization.append(space_utilization)
            problem.available_at.append(randint(0,5))
            problem.sailing_costs.append(fuel_costs)

        problem.vessels += vessel_diff

        print_problem(problem)


#problem = AsiaToEuropeProblem()
#correct_port_frequency(problem)
#populate(problem)
#problem.lock_x = get_x(problem, True)

#print_problem(problem)
#save_instance(problem)

def create_dir(dir):
    if not os.path.exists("Output/" + dir):
        os.makedirs("Output/" + dir)

    if not os.path.exists("Instances/" + dir):
        os.makedirs("Instances/" + dir)

def increase_total_demand(inc=1.02, steps = 25, vessels = 4):

    perc = 1.0

    problem = AsiaToEuropeProblem()
    problem.vessels = vessels
    correct_port_frequency(problem)
    populate(problem)

    dir_name = "total_demand_two/" + str(problem.case_name) + "_" + str(problem.contracts) + "_" + str(problem.voyages) \
               + "_" + str(problem.vessels) + "_" + str(problem.products) + "_" + str(1.0)
    create_dir(dir_name)

    print_problem(problem, dir=dir_name)

    for step in range(steps):

        for contract_key in range(problem.contracts):

            contract = problem.contracts_load[contract_key]

            new_contract = [(int(p[0]*inc), int(p[1]), int(p[2]*inc)) for p in contract]

            problem.contracts_load[contract_key] = new_contract

        perc = float("{0:.4f}".format(round(perc*inc,2)))

        dir_name = "total_demand_two/" + str(problem.case_name) + "_" + str(problem.contracts) + "_" + str(problem.voyages)\
                                        + "_" + str(problem.vessels) + "_" + str(problem.products) + "_" + str(perc)
        create_dir(dir_name)

        print_problem(problem, dir=dir_name)

'''problem = AsiaToEuropeProblem()
populate(problem)

l = []

for i, c in enumerate(problem.contracts_info):
    l_el = (c[-1]/(c[1] - c[0]), i, c)
    if c[1] - c[0] > 4:
        l.append(l_el)

print(sorted(l))'''

increase_total_demand(vessels=3)
